#############################################################################
# xiron_plugin.tcl Copyright 2006, 2007, 2008, 2009 r0lZ
# PgcEdit plugin to fix and modify DVDs created by a XiRON recorder.
#
# This plugin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This plugin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################
# Usage: copy this file in the <PgcEdit's install dir>/plugins directory.
# A "XiRON" menu will be added in the Plugins menu.
# You need PgcEdit 0.6.0 or more to use this plugin.
# It is better to use v 0.6.1 or more.
#############################################################################
# Version History:
#
# v 1.0 (October 6, 2006)
# - First public release
# v 1.1 (May 25, 2007)
# - Little bug fixed: the PGC and GPRM user comments (if any) were not saved
#   when the DVD was saved during Fix VMGM VOB IDs.
# v 1.2 (June 6, 2007)
# - The temporary INI file needed to launch VobBlanker is now created with an
#   unique filename to ensure that no conflict exists with another instance
#   of PgcEdit, and is automatically deleted after use.
#############################################################################


# Add Plugins -> xiron submenu
if {! [winfo exists .mainmenu.plugins.xiron]} {
	.mainmenu.plugins add cascade -label "XiRON" -menu .mainmenu.plugins.xiron
	menu .mainmenu.plugins.xiron

	.mainmenu.plugins.xiron add command -label "Fix all known problems"  \
			-command {::plugins::xiron::fix_all}

	.mainmenu.plugins.xiron add separator
	.mainmenu.plugins.xiron add command -label "1. Fix VMGM VOB IDs"  \
			-command {::plugins::xiron::fix_vids}
	.mainmenu.plugins.xiron add command -label "2. Fix Titles streams attributes"  \
			-command {::plugins::xiron::fix_streams}
	.mainmenu.plugins.xiron add command -label "3. Fix Titles LBA Pointers"  \
			-command {::plugins::xiron::fix_lbas}
	.mainmenu.plugins.xiron add command -label "4. FixVTS -p"  \
			-command {::plugins::xiron::fix_vts}
	.mainmenu.plugins.xiron add command -label "5. Build Search pointers with VobBlanker"  \
			-command {::plugins::xiron::build_srptrs}
	.mainmenu.plugins.xiron add command -label "6. Build Time Maps"  \
			-command {::plugins::xiron::build_tmaps}

	.mainmenu.plugins.xiron add separator
	.mainmenu.plugins.xiron add command -label "Enhance authoring"  \
			-command {::plugins::xiron::enhance}
	.mainmenu.plugins.xiron add command -label "Redo authoring"  \
			-command {::plugins::xiron::redo}

	.mainmenu.plugins.xiron add separator
	.mainmenu.plugins.xiron add command -label "About" -command {
			tk_messageBox  -title "XiRON plugin" -icon info \
					-message "XiRON plugin for PgcEdit 7.4 or more.\nVersion $::plugins::xiron::version\nAuthor: r0lZ\nFreeware!"
		}
}

namespace eval ::plugins {}
namespace eval ::plugins::xiron {

############## User modifiable constants #############
# Path to VobBlanker executable
set vobblanker_exe "C:\\Program Files\\VobBlanker\\VobBlanker.exe"
# Default output folder for VobBlanker's process.  Leave empty ("") for "Use input folder".
set vobblanker_target "D:\\rip\\tmp"
########## End of user modifiable constants ##########

set version "1.2"

# Fix VOB IDs in VIDEO_TS.VOB
proc fix_vids {} {
	set title "XiRON plugin - Fix VMGM VOB IDs"
	if {![info exists ::dvddir]} {
		show_msg "No DVD loaded!"
		bell
		return -1
	}
	if {$::dvd_modified} {
		set txt "Unsaved changes will be lost!"
		set txt "$txt\nContinue anyway?"
		set r [tk_messageBox -message $txt -title $title -icon warning \
				-type okcancel -default cancel]
		if {$r == "cancel"} {
			return -1
		}
	}

	show_msg "Fixing VOB IDs in VMGM..."
	set filename [file join $::dvddir "VIDEO_TS.VOB"]
	if {![file exists $filename]} {
		show_msg "VIDEO_TS.VOB not found!"
		bell
		return -1
	}

	set numsectors [expr [file size $filename]/2048]
	set fh [::menued::openbinfile $filename r+]
	if {$fh == ""} {
		show_msg "Cannot open VIDEO_TS.VOB!"
		bell
		return -1
	}

	set sector 0
	set numfixed 0
	while {1} {
		set navpack [::menued::load_nav_pack $filename $fh $sector]
		if {$navpack == ""} {
			show_msg "Nav Pack at sector $sector is invalid (in VIDEO_TS.VOB)!"
			bell
			close $fh
			return -1
		}
		set vid [get_int navpack 0x41F 2]
		if {$vid == 1} {
			show_msg "Nav Pack at sector $sector has already a VOB ID = 1!  (VIDEO_TS.VOB already patched or wrong DVD?)"
			bell
			close $fh
			return -1
		}
		set new_vid [expr $vid-1]
		set_int navpack 0x41F $new_vid 2
		if {! [::menued::save_nav_pack $filename $fh $sector $navpack]} {
			show_msg "Cannot save nav_pack at LBA $sector in VIDEO_TS.VOB!"
			bell
			close $fh
			return -1
		}
		incr numfixed

		set end_of_vobu [get_int navpack 0x040F 4]
		if {$end_of_vobu == 0} {
			break
		}
		set sector [expr $sector + $end_of_vobu + 1]
		if {$sector >= $numsectors} {
			break
		}
	}
	close $fh
	::main::ReRead_DVD false true
	show_msg "VMGM VOB IDs fixed OK.  $numfixed nav packs fixed."
	return $numfixed
}

# Fix Titles LBA Pointers
proc fix_lbas {} {
	set title "XiRON plugin - Fix Titles LBA Pointers"
	if {![info exists ::dvddir]} {
		show_msg "No DVD loaded!"
		bell
		return -1
	}
	if {$::dvd_modified} {
		set txt "Unsaved changes will be lost!"
		set txt "$txt\nContinue anyway?"
		set r [tk_messageBox -message $txt -title $title -icon warning \
				-type okcancel -default cancel]
		if {$r == "cancel"} {
			return -1
		}
	}
	set error [get_vobu_lbas 1 lbas]
	if {$error != 0} {
		return -1
	}

	show_msg "Patching IFO PGCs..."
	for {set pgc 1} {$pgc <= $::pgcs(1,numpgcs)} {incr pgc} {
		set numcells $::pgcs(1,$pgc,numcells)
		set postable $::pgcs(1,$pgc,cellpostable)
		for {set c 1} {$c <= $numcells} {incr c} {
			set cellposidx [expr {($c-1)*4}]
			set vid [get_int postable $cellposidx 2]
			set cid [get_int postable [expr {$cellposidx+3}] 1]
			# Replaces the dummy, inexistant cell 1/255 by cell 1/1
			if {$vid==1 && $cid==255 && ![info exists lbas($vid,$cid,entry)]} {
				set cid 1
				set_int postable [expr {$cellposidx+3}] $cid 1
				set ::pgcs(1,$pgc,cellpostable) $postable
			}

			set cellpbidx [expr {($c-1)*24}]
			if {[info exists lbas($vid,$cid,entry)]} {
				set_int ::pgcs(1,$pgc,cellpbtable) [expr $cellpbidx+8]  $lbas($vid,$cid,entry) 4
				set_int ::pgcs(1,$pgc,cellpbtable) [expr $cellpbidx+16] $lbas($vid,$cid,laststart) 4
				set_int ::pgcs(1,$pgc,cellpbtable) [expr $cellpbidx+20] $lbas($vid,$cid,lastend) 4
			}
		}
	}

	show_msg "Patching VTS_C_ADT..."
	set table $::raw_sectors(1,VTS_C_ADT)
	set len [string length $table]
	set numfixed 0
	for {set idx 8} {$idx < $len} {incr idx 12} {
		set vid [get_int table $idx 2]
		set cid [get_int table [expr {$idx+2}] 1]
		if {[info exists lbas($vid,$cid,entry)]} {
			if {[get_int table [expr {$idx+4}] 4] != $lbas($vid,$cid,entry) || \
				[get_int table [expr {$idx+8}] 4] != $lbas($vid,$cid,lastend)} {
					set_int table [expr {$idx+4}] $lbas($vid,$cid,entry) 4
					set_int table [expr {$idx+8}] $lbas($vid,$cid,lastend) 4
					incr numfixed
			}
		}
	}
	set ::raw_sectors(1,VTS_C_ADT) $table
	::main::dvd_modified true false
	show_msg "VTS1 Done.  Fixed bad VOBU addresses of $numfixed cells."
}

proc build_tmaps {} {
	if {![info exists ::dvddir]} {
		show_msg "No DVD loaded!"
		bell
		return -1
	}
	# rebuild the time maps
	show_msg "Building Time Maps..."
	::tmap::rebuild_tmap_from_vobs 1
	::tmap::dont_need_new_tmap 1

	# Remove Time Search PUO in Title Play Map table
	show_msg "Removing Time Search PUOs in Title Play Map table."
	set table $::raw_sectors(0,VMG_TT_SRPT)
	set numtitles [get_int table 0 2]
	for {set t 0} {$t < $numtitles} {incr t} {
		set idx [expr 8+($t*12)]
		set_int table $idx [expr [get_int table $idx 1] & 0xFC] 1
	}
	set ::raw_sectors(0,VMG_TT_SRPT) $table
	::main::dvd_modified true false
}

proc fix_streams {} {
	set default_lang_code "fr"	;# User modifiable constant

	if {![info exists ::dvddir]} {
		show_msg "No DVD loaded!"
		bell
		return -1
	}
	show_msg "Fixing streams."
	set mat_table $::raw_sectors(1,VTSI_MAT)
	set numaudios  [get_int mat_table [expr {0X200+2}] 2]
	set numsubpics [get_int mat_table [expr {0X200+0X54}] 2]
	if {$numaudios != 1 || $numsubpics > 1} {
		show_msg "Number of audio and/or subpic streams is not 1!  Wrong DVD?"
		bell
		return -1
	}

	# Source picture already letterboxed?
	set byte1 [get_int mat_table 0x201 1]
	if {[expr $byte1 & 4] > 0} {
		set default "yes"
	} else {
		set default "no"
	}
	set txt "Source picture already letterboxed?"
	set txt "$txt\n(Answer YES if the picture has horizontal black borders.)"
	set txt "$txt\nCurrent value: $default"
	set r [tk_messageBox -message $txt -title "XiRON Plugin - Fix streams" \
			-icon question -type yesnocancel -default $default]
	if {$r == "cancel"} {
		show_msg "Video already letterboxed change skipped!"
	} else {
		show_msg "Setting video already letterboxed flag to: $r"
		if {$r == "yes"} {
			set byte1 [expr $byte1 | 4]
		} else {
			set byte1 [expr $byte1 & 0xFB]
		}
		set_int mat_table 0x201 $byte1 1
	}

	# Get AUDIO lang code
	set aattr [string range $mat_table [expr 0X200+4] [expr 0X200+11]]
	set lang_code [lindex [::streams::audio_attributes $aattr 0] 0]
	set orig_lang_code $lang_code
	if {$lang_code == "--"} {set lang_code $default_lang_code}
	show_msg "Original audio language code: $orig_lang_code"

	set txt "Enter Audio lang code"
	set txt "$txt\n2 lower case letters (\"en\", \"fr\"...)"
	set txt "$txt\nor empty string to remove the code."
	set txt "$txt\nCurrent value: $orig_lang_code"
	set r [::stringreq::stringreq -window .xiron_stream -parent . \
			-title "XiRON Plugin - Fix streams" -text $txt \
			-def $lang_code -callback ::streams::validate_langcode \
			-width 2 -geometry "c" -font $::config(fixed_width_font) \
			-cancelreturn "cancel"]
	if {$r == "cancel"} {
		show_msg "Audio lang code change skipped!"
	} else {
		if {$r == ""} {set r \000\000}
		if {$r == $orig_lang_code || [string length $r] != 2} {
			show_msg "Audio language code not changed."
		} else {
			show_msg "Audio language code changed to: $r"
			set mat_table [string range $mat_table 0 517]$r[string range $mat_table 520 end]
			set byte0 [get_int mat_table [expr 0X200+4] 1]
			if {$r == "\000\000"} {
				set_int mat_table [expr 0x200+4] [expr $byte0 & 0xF3] 1
			} else {
				set_int mat_table [expr 0x200+4] [expr $byte0 | 0x04] 1
			}
		}
	}

	# Remove subpic definition from mat table
	show_msg "Removing subpic stream..."
	set_int mat_table [expr 0X200+0x54] 0 2
	set mat_table [string range $mat_table 0 [expr 0X200+0x55]]\000\000\000\000\000\000[string range $mat_table [expr 0X200+0x5C] end]
	# and remove stream assignments in PGCs
	for {set pgc 1} {$pgc <= $::pgcs(1,numpgcs)} {incr pgc} {
		set ::pgcs(1,$pgc,header) \
			[string range $::pgcs(1,$pgc,header) 0 27][string repeat \000 128][string range $::pgcs(1,$pgc,header) 156 end]
	}

	# Fix NTSC/PAL standard in VTSM domain
	if {[expr [get_int mat_table 0x200 1] & 0x30] == 16} {
		show_msg "Forcing PAL for VTSM 1"
		set mode [expr [get_int mat_table 0x100 1] | 0x10]
		set_int mat_table 0x100 $mode 1
	}


	set ::raw_sectors(1,VTSI_MAT) $mat_table
	::main::dvd_modified true false
	show_msg "Done.  VTSI_MAT table and PGCs fixed."
	return 0
}

proc fix_vts {{start false}} {
	if {![info exists ::dvddir]} {
		show_msg "No DVD loaded!"
		bell
		return -1
	}
	::main::display_pgc pgc 1 1
	::toolscfg::fixvts $start true
}

proc build_srptrs {} {
	if {![info exists ::dvddir]} {
		show_msg "No DVD loaded!"
		bell
		return -1
	}
	show_msg "Running VobBlanker..."
	# launch VobBlanker
	# Retrieve VobBlanker.exe filename
	if {! [file exists $::plugins::xiron::vobblanker_exe]} {
		show_msg "VobBlanker not configured.  Please edit xiron_plugin.tcl manually!"
		bell
		return false
	}
	# Generate INI file for VobBlanker
	set out "\[General\]\n"
	set out "${out}InputPath=[file nativename $::dvddir]\n"
	if {$::plugins::xiron::vobblanker_target == ""} {
		set target [file nativename $::dvddir]
		set check_folder 1
	} else {
		set target $::plugins::xiron::vobblanker_target
		set check_folder 0
	}
	set out "${out}OutputPath=$target\n"
	set out "${out}CheckMenu=1\n"
	set out "${out}CheckTitle=1\n"
	set out "${out}CheckIfoPuops=0\n"
	set out "${out}CheckVobPuops=0\n"
	# may be needed when a V/CID has been changed in the VOBs
	set out "${out}CheckFolder=$check_folder\n"

	# Options moved in subsections since VB 2.1,
	# but kept here also for compatibility with older versions
	#
	# In PGC options section:
	set out "${out}CheckPostPre=0\n"
	set out "${out}CheckPostPreForced=0\n"
	set out "${out}CheckNextMenu=0\n"
	set out "${out}CheckNextTitle=0\n"
	set out "${out}CheckPrevMenu=0\n"
	set out "${out}CheckPrevTitle=0\n"
	set out "${out}CheckAttrTitle=0\n"
	# In FixVOBPointers section:
	set out "${out}CheckFixPointers=1\n"

	# New section in VB 2.1
	set out "${out}\[PGC options\]\n"
	set out "${out}CheckPostPre=0\n"
	set out "${out}CheckPostPreForced=0\n"
	set out "${out}CheckNextMenu=0\n"
	set out "${out}CheckNextTitle=0\n"
	set out "${out}CheckPrevMenu=0\n"
	set out "${out}CheckPrevTitle=0\n"
	set out "${out}CheckAttrTitle=0\n"

	# New section in VB 2.1
	set out "${out}\[FixVOBPointers\]\n"
	set out "${out}CheckFixTitlePointers=1\n"
	set out "${out}CheckFixMenuPointers=1\n"

	set out "${out}\[ScanVOBUs\]\n"
	set out "${out}CheckAutoScan=1\n"

	set out "${out}\[VobLayout\]\n"
	set out "${out}CheckVobFill=0\n"
	set out "${out}CheckVobOrder=1\n"

	set out "${out}CheckVobMUnref=0\n"
	set out "${out}CheckVobLUnref=0\n"
	set out "${out}CheckVobRenumber=1\n"
	set out "${out}CheckClockDisc=1\n"

	set out "${out}\[VIDEO_TS\]\n"
	set out "${out}ActionVTS=PROCESS\n"

	set out "${out}\[VTS01\]\n"
	set out "${out}ActionVTS=PROCESS\n"
	#set out "${out}\[Backup\]\n"
	#set out "${out}CheckBckVts=1\n"

	# Write INI file
	set inifilename [file join $::config(tempdir) "VobBlanker_xiron_[pid].INI"]
	if {[saveFile $inifilename auto $out] != 0} {
		return false
	}
	# launch VobBlanker
	wm iconify .
	set rc [catch {exec -- $::plugins::xiron::vobblanker_exe $inifilename} errstr]
	wm deiconify .
	catch {file delete $inifilename}
	if {$rc} {
		tk_messageBox -title "VobBlanker error" -message $errstr -icon error
		return false
	} else {
		return true
	}
}

proc fix_all {} {
	if {![info exists ::dvddir]} {
		show_msg "No DVD loaded!"
		bell
		return -1
	}
	fix_vids

	fix_streams
	show_msg "Saving DVD..."
	::main::Write_DVD false

	fix_lbas
	show_msg "Saving DVD..."
	::main::Write_DVD false

	show_msg "Running FixVTS -p"
	fix_vts true

	build_tmaps
	show_msg "Saving DVD..."
	::main::Write_DVD false

	# finish with VobBlanker process
	build_srptrs
}

proc get_vobu_lbas {vts lbas_upvar} {
	upvar lbas $lbas_upvar
	if {![info exists ::dvddir]} {bell; return -1}

	set error 0

	# compute number of sectors to process for progress bar
	show_msg "Reading Cells VOBU LBAs..."
	set numsectors 0
	for {set i 1} {$i < 10} {incr i} {
		set filename [file join $::dvddir "VTS_[format %02d $vts]_$i.VOB"]
		if {[file exists $filename]} {
			incr numsectors [expr [file size $filename] / 2048]
		} else {
			break
		}
	}

	set vobnum 1
	set lba 0
	set sector 0
	set lastlba -1
	set vid -1
	set cid -1
	set filename [file join $::dvddir "VTS_[format %02d $vts]_$vobnum.VOB"]
	if {![file exists $filename]} {
		show_msg "$filename not found!"
		bell
		return -1
	}
	#show_msg "Opening file [file tail $filename]"
	set vobsize [expr [file size $filename] / 2048]
	show_msg "Reading Cells VOBU LBAs in [file tail $filename]..." $lba $numsectors
	set fh [::menued::openbinfile $filename r+]
	if {$fh != ""} {
		while {1} {
			set navpack [::menued::load_nav_pack $filename $fh $sector]
			if {$navpack == ""} {
				show_msg "Nav Pack at sector $lba is invalid (in [file tail $filename] at sector $sector)!" $lba $numsectors

				set l [search_next_valid_nav_pack $filename $fh $sector]
				set lba [expr $lba + ([lindex $l 0]-$sector)]
				set sector  [lindex $l 0]
				set navpack [lindex $l 1]
				if {$navpack == ""} {
					if {$sector < $vobsize} {
						set error -1
					}
					break
				}
				show_msg "A new valid nav pack was found at sector $sector.  Continuing..." $lba $numsectors
			}
			set vid [get_int navpack 0x41F 2]
			set cid [get_int navpack 0x422 1]
			#puts "$vid/$cid at LBA $lba (in [file tail $filename] at $sector)"
			if {![info exists lbas($vid,$cid,entry)]} {
				show_msg "Found VCID $vid/$cid at LBA $lba (in [file tail $filename] at sector $sector)" $lba $numsectors
				set lbas($vid,$cid,entry) $lba
				if {$lastlba != -1} {
					set lbas($lastvid,$lastcid,laststart) $lastlba
					set lbas($lastvid,$lastcid,lastend) [expr $lba-1]
				}
				set lastvid $vid
				set lastcid $cid
			}

			set end_of_vobu [get_int navpack 0x040F 4]
			# XiRION bug: wrong VOBU end address
			if {$cid == 255} {set end_of_vobu 0}
			set sector [expr $sector + $end_of_vobu + 1]
			set lastlba $lba
			set lba    [expr $lba    + $end_of_vobu + 1]
			if {$lba >= $numsectors || $sector < 0} {
				break
			}
			if {$sector >= $vobsize} {
				close $fh
				set fh ""
				incr vobnum
				set filename [file join $::dvddir "VTS_[format %02d $vts]_$vobnum.VOB"]
				set sector [expr $sector - $vobsize]
				if {[file exists $filename]} {
					#show_msg "Opening file [file tail $filename]"
					set vobsize [expr [file size $filename] / 2048]
					show_msg "Reading Cells VOBU LBAs in [file tail $filename]..." $lba $numsectors
					set fh [::menued::openbinfile $filename r+]
					if {$fh == ""} break
				} else {
					break
				}
			}
		}
		if {$fh != ""} {close $fh}
		if {$lastlba != -1} {
			set lbas($lastvid,$lastcid,laststart) $lastlba
			set lbas($lastvid,$lastcid,lastend) [expr $lba-1]
		}
		show_msg "Done fixing VOBU LBAs of VTST $vts."
	} else {
		set error -1
	}
	return $error
}

proc search_next_valid_nav_pack {filename fh sector} {
	set vobsize [expr [file size $filename] / 2048]
	incr sector
	if {$sector >= $vobsize} {return [list $sector ""]}
	set navpack [::menued::load_nav_pack $filename $fh $sector]
	while {$navpack == ""} {
		incr sector
		if {$sector >= $vobsize} {return [list $sector ""]}
		set navpack [::menued::load_nav_pack $filename $fh $sector]
	}
	return [list $sector $navpack]
}


# Enhance authoring
proc enhance {} {
	if {![info exists ::dvddir]} {bell; return}
	# how many real Titles?
	set numtitles [number_of_real_titles]
	if {$numtitles == -1} return

	# return to menu after playing a regular title
	for {set p 1} {$p <= $numtitles} {incr p} {
		set_type_cmds post ::pgcs(1,$p) {{30 08 00 01 01 C0 00 00}}
	}
	# select Title 1 button when last title has finished playing
	foreach p [list $numtitles [expr $numtitles*2]] {
		set ::pgcs(1,$p) {{* pre} {71 00 00 04 00 01 00 00} {30 08 00 01 01 C0 00 00} {* post} {30 08 00 01 01 C0 00 00} {* cell}}
	}
	# ... and modify the "DVD" button to jump to the first title in loop mode
	if {[info exists ::buttons::butsinfo(menu,0,1,1,1,1,command)] && \
			$::buttons::butsinfo(menu,0,1,1,1,1,command) == "20 01 00 00 00 00 04 05"} {
		set ::buttons::butsinfo(menu,0,1,1,1,1,command) "30 02 00 00 00 [format %02X [expr $numtitles+1]] 00 00"
		set ::buttons::butsinfo(menu,edited) 1
		set ::buttons::butsinfo(menu,0,edited) 1
		set ::buttons::butsinfo(menu,0,1,edited) 1
		set ::buttons::butsinfo(menu,0,1,1,edited) 1
	}

	# Move Title menu from VMGM PGC 1 to 2,
	# so that pressing TITLE calls now the next title.
	# ROOT still calls the menu.
	::utils::set_menu_type 0 1 2 2

	# set NextPGCN link to itself
	for {set p 1} {$p <= $::pgcs(1,numpgcs)} {incr p} {
		set_int ::pgcs(1,$p,header) 156 $p 2
	}

	# Change color of button highlights
	set_int ::menus(0,1,1,header) 0xB1 0xE0748600 4

	# set provider ID
	set newid "XIRION_PGCEDIT"
	set newid "$newid[string repeat \000 [expr {32-[string length $newid]}]]"
	set ::raw_sectors(0,VMGM_MAT) \
			[string replace $::raw_sectors(0,VMGM_MAT) 0X40 0X5F $newid]

	::main::dvd_modified true false
	::main::display_pgc pgc 0 0
	show_msg "Authoring enhanced successfully.  Please save the DVD!"

	# input DVD name
	::utils::set_General_Name_GUI
}

# Redo authoring
proc redo {} {
	if {![info exists ::dvddir]} {bell; return}
	# how many real Titles?
	set numtitles [number_of_real_titles]
	if {$numtitles == -1} return

	#set ::warning_delete_last_pgc true
	for {set p $::pgcs(1,numpgcs)} {$p >= $numtitles} {incr p -1} {
		::main::display_pgc pgc 1 1
		::utils::delete_last_pgc true
	}
	#set ::warning_delete_last_pgc false
	::utils::delete_all_DVD_commands true

	# we don't need the dummy title
	incr numtitles -1

	# set NextPGCN link to itself
	for {set p 1} {$p <= $::pgcs(1,numpgcs)} {incr p} {
		set_int ::pgcs(1,$p,header) 156 $p 2
	}

	##### Build commands #####
	### FP-PGC
	set ::pgcs(0,0) {
		{* pre}
		{71 00 00 00 FF FF 00 00}
		{30 02 00 00 00 01 00 00}
		{* post}
		{* cell}
	}

	### VMGM 1
	set tmp {
		{00 B1 00 00 FF FF 00 04}
		{71 00 00 00 00 00 00 00}
		{20 06 00 00 00 00 04 01}
		{00 00 00 00 00 00 00 00}
		{71 00 00 01 00 01 00 00}
		{00 00 00 00 00 00 00 00}
		{61 00 00 0F 00 84 00 00}
		{74 00 00 0F 00 01 00 00}
		{77 00 00 0F 00 03 00 00}
		{73 00 00 0F 00 01 00 00}
		{75 00 00 0F 04 00 00 00}
		{46 00 00 00 00 0F 00 00}
		{00 00 00 00 00 00 00 00}
		{61 00 00 0F 00 84 00 00}
		{74 00 00 0F 00 01 00 00}
		{76 00 00 0F 00 03 00 00}
		{73 00 00 0F 00 02 00 00}
	}
	for {set i 2} {$i <= [expr ($numtitles/3)+2]} {incr i} {
		lappend tmp "20 A6 00 0F 00 [format %02X $i] 00 [format %02X $i]"
	}
	lappend tmp {00 00 00 00 00 00 00 00} {56 00 00 00 04 00 00 00}
	set_type_cmds pre  ::menus(0,1,1) $tmp
	set_type_cmds post ::menus(0,1,1) {
		{71 00 00 01 00 00 00 00}
		{20 A1 00 02 00 01 00 10}
		{30 02 00 00 00 01 00 00}
	}
	# Enable RESUME with the GoUp key
	set_int ::menus(0,1,1,header) 0xA0 0xFFFF 2
	# Change color of button highlights
	set_int ::menus(0,1,1,header) 0xB1 0xE0748600 4
	# Modify the "DVD" button to enable loop mode: LinkTailPGC
	if {[info exists ::buttons::butsinfo(menu,0,1,1,1,1,command)]} {
		set ::buttons::butsinfo(menu,0,1,1,1,1,command) {20 01 00 00 00 00 00 0D}
		set ::buttons::butsinfo(menu,edited) 1
		set ::buttons::butsinfo(menu,0,edited) 1
		set ::buttons::butsinfo(menu,0,1,edited) 1
		set ::buttons::butsinfo(menu,0,1,1,edited) 1
	}
	# Modify the last (dummy) Title button command to go to main DVD menu page
	set but [expr ($numtitles%3)+1]
	set vid [expr ($numtitles/3)+2]
	if {[info exists ::buttons::butsinfo(menu,0,$vid,1,1,$but,command)]} {
		set ::buttons::butsinfo(menu,0,$vid,1,1,$but,command) {20 06 00 00 00 00 04 01}
		set ::buttons::butsinfo(menu,0,$vid,edited) 1
		set ::buttons::butsinfo(menu,0,$vid,1,edited) 1
	}

	### VMGM 2 (Title menu)

	# Move Title menu from VMGM PGC 1 to 2,
	# so that pressing TITLE calls now the next title.
	# ROOT still calls the menu.
	::utils::set_menu_type 0 1 2 2

	set tmp {
		{71 00 00 01 00 00 00 00}
		{61 00 00 0F 00 84 00 00}
		{73 00 00 0F 00 01 00 00}
	}
	lappend tmp	"00 E1 00 0F 00 [format %02X $numtitles] 00 06" \
			{20 04 00 00 00 00 00 01} \
			{00 00 00 00 00 00 00 00} \
			{71 00 00 0E 00 00 00 00}
	for {set i 1} {$i <= $numtitles} {incr i} {
		lappend tmp {73 00 00 0E 00 01 00 00} "30 22 00 00 00 [format %02X $i] 0E 0F"
	}
	set_type_cmds pre ::menus(0,1,2) $tmp

	### VMGM 3
	set_type_cmds pre ::menus(0,1,3) {
		{20 A4 00 01 00 00 00 02}
		{20 04 00 00 00 00 00 01}
	}

	### VTST 1 (Root menu)
	set_type_cmds pre ::menus(1,1,1) {
		{71 00 00 01 00 01 00 00}
		{30 06 00 01 00 C0 00 00}
	}

	### Titles
	for {set p 1} {$p <= $numtitles} {incr p} {
		set_type_cmds pre  ::pgcs(1,$p) {
			{71 00 00 02 00 01 00 00}
		}
		set_type_cmds post ::pgcs(1,$p) {
			{71 00 00 02 00 00 00 00}
			{30 08 00 03 01 C0 00 00}
		}
	}
	# Overwrite Title 1 pre-commands
	set_type_cmds pre ::pgcs(1,1) {
		{00 B1 00 00 FF FF 00 03}
		{30 08 00 01 01 C0 00 00}
		{00 00 00 00 00 00 00 00}
		{71 00 00 02 00 01 00 00}
	}

	##### Misc #####
	# set provider ID
	set newid "XIRION_PGCEDIT"
	set newid "$newid[string repeat \000 [expr {32-[string length $newid]}]]"
	set ::raw_sectors(0,VMGM_MAT) \
			[string replace $::raw_sectors(0,VMGM_MAT) 0X40 0X5F $newid]

	::main::dvd_modified true false
	::main::display_pgc pgc 0 0
	show_msg "Authoring remade successfully.  Please save the DVD!"

	# input DVD name
	::utils::set_General_Name_GUI
}

proc number_of_real_titles {} {
	if {! [info exists ::menus(0,1,2)]} {
		show_msg "Missing VMGM or incorrect format!  Wrong DVD?"
		bell
		return -1
	}
	set numtitles 0
	set cmds $::menus(0,1,2)
	set cmds [lsearch -all -inline -regexp $cmds "30 22 00 00 00 .. 04 01"]
	if {[llength $cmds] == 0} {
		show_msg "VMGM is not of correct format!  Wrong DVD?"
		bell
		return -1
	}
	set numtitles [expr 0x[lindex [lindex $cmds end] 5]]
	if {$numtitles != [expr $::pgcs(1,numpgcs)/2]} {
		show_msg "Cannot determine the number of real Titles!  Wrong DVD?"
		bell
		return -1
	}
	return $numtitles
}

}	;# end of namespace ::plugins::xiron
